# -*- coding: utf-8 -*-
# Generated by Django 1.11.3 on 2017-07-12 07:59
from __future__ import unicode_literals

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("tests", "0004_user_is_dead"),
    ]

    operations = [
        migrations.CreateModel(
            name="Car",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),  # noqa
                ("name", models.CharField(max_length=60)),
            ],
        ),
        migrations.CreateModel(
            name="Country",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),  # noqa
                ("name", models.CharField(max_length=60)),
                ("short_name", models.CharField(max_length=30)),
            ],
        ),
        migrations.CreateModel(
            name="Part",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),  # noqa
                ("name", models.CharField(max_length=60)),
                (
                    "car",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, to="tests.Car"
                    ),
                ),  # noqa
                (
                    "country",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, to="tests.Country"
                    ),
                ),  # noqa
            ],
        ),
        migrations.AddField(
            model_name="car",
            name="country",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, to="tests.Country"
            ),  # noqa
        ),
    ]
