"""Dynamic REST (or DREST) is an extension of Django REST Framework.

DREST offers the following features on top of the standard DRF kit:

- Linked/embedded/side-loaded relationships
- Field inclusions/exclusions
- Field-based filtering/sorting
- Directory panel for the browsable API
- Optimizations
"""
__version__ = "2.4.0"
default_app_config = "dynamic_rest.apps.DynamicRestConfig"
