# -*- coding: utf-8 -*-
# Generated by Django 1.9 on 2016-03-10 10:52
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("tests", "0001_initial"),
    ]

    operations = [
        migrations.AddField(
            model_name="user",
            name="date_of_birth",
            field=models.DateField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name="group",
            name="name",
            field=models.TextField(unique=True),
        ),
    ]
