.main ul.breadcrumb {
    margin-top: 0px;
}

@media screen and (max-width: 1800px) {
    .directory {
        display: none;
    }
}

.directory {
    font-family: 'Open Sans', 'Helvetica Neue', Helvetica, sans-serif;
    position: fixed;
    left: 0;
    top: 50px;
    bottom: 0;
    overflow: auto;
    min-width: 320px;
    padding-top: 20px;
    background-color: #222;
}

.directory ul {
    margin: 0;
    padding: 0;
}

.directory li {
    margin: 0;
    padding-left: 10px;
    list-style-type: none;
}

.directory li a:hover {
    color: white;
    text-decoration: none;
}

.directory li ul {
    padding-left: 20px;
}

a {
    text-decoration: none;
}

.directory a, .directory li {
    display: block;
    font-size: 16px;
    color: rgba(255, 255, 255, .60);
}

.directory a.active, .directory li.active {
    color: white;
}

html, body {
    margin: 0;
    padding: 0;
    height: 100%;
}
